@@ -4,11 +4,10 @@ use crate::updates::decoder::{Decode, Decoder};
 use crate::updates::encoder::{Encode, Encoder};
 use crate::utils::client_hasher::ClientHasher;
 use crate::*;
-use std::cell::UnsafeCell;
 use std::collections::hash_map::Entry;
 use std::collections::HashMap;
 use std::hash::BuildHasherDefault;
-use std::ops::Index;
+use std::ops::Deref;
 use std::rc::Rc;
 use std::vec::Vec;
 
@@ -134,9 +133,46 @@ impl Encode for StateVector {
     }
 }
 
+/// Snapshot describes a state of a document store at a given point in (logical) time.
+#[derive(Default, Debug, Clone, PartialEq, Eq)]
+pub struct Snapshot {
+    /// Compressed information about all deleted blocks at current snapshot time.
+    pub delete_set: DeleteSet,
+    /// Logical clock describing a current snapshot time.
+    pub state_map: StateVector,
+}
+
+impl Snapshot {
+    pub fn new(state_map: StateVector, delete_set: DeleteSet) -> Self {
+        Snapshot {
+            state_map,
+            delete_set,
+        }
+    }
+
+    pub(crate) fn is_visible(&self, id: &ID) -> bool {
+        self.state_map.contains(id) && !self.delete_set.is_deleted(id)
+    }
+}
+
+impl Encode for Snapshot {
+    fn encode<E: Encoder>(&self, encoder: &mut E) {
+        self.delete_set.encode(encoder);
+        self.state_map.encode(encoder);
+    }
+}
+
+impl Decode for Snapshot {
+    fn decode<D: Decoder>(decoder: &mut D) -> Self {
+        let ds = DeleteSet::decode(decoder);
+        let sm = StateVector::decode(decoder);
+        Snapshot::new(sm, ds)
+    }
+}
+
 /// A resizable list of blocks inserted by a single client.
 pub(crate) struct ClientBlockList {
-    list: Vec<UnsafeCell<block::Block>>,
+    list: Vec<Box<block::Block>>,
     integrated_len: usize,
 }
 
@@ -182,24 +218,14 @@ impl ClientBlockList {
         }
     }
 
-    pub fn get(&self, index: usize) -> &Block {
-        let ptr = self.list[index].get();
-        unsafe { &*ptr }
+    pub fn try_get(&self, index: usize) -> Option<BlockPtr> {
+        let block_ref = self.list.get(index)?;
+        Some(BlockPtr::from(block_ref))
     }
 
-    pub fn get_mut(&self, index: usize) -> &mut Block {
-        let ptr = self.list[index].get();
-        unsafe { &mut *ptr }
-    }
-
-    pub fn try_get(&self, index: usize) -> Option<&Block> {
-        let ptr = self.list.get(index)?.get();
-        Some(unsafe { &*ptr })
-    }
-
-    pub fn try_get_mut(&self, index: usize) -> Option<&mut Block> {
-        let ptr = self.list.get(index)?.get();
-        Some(unsafe { &mut *ptr })
+    pub fn get(&self, index: usize) -> BlockPtr {
+        let block_ref = &self.list[index];
+        BlockPtr::from(block_ref)
     }
 
     /// Gets the last clock sequence number representing the state of inserts made by client
@@ -218,29 +244,17 @@ impl ClientBlockList {
     /// Returns first block on the list - since we only initialize [ClientBlockList]
     /// when we're sure, we're about to add new elements to it, it always should
     /// stay non-empty.
-    pub(crate) fn first(&self) -> &Block {
+    pub(crate) fn first(&self) -> BlockPtr {
         self.get(0)
     }
 
     /// Returns last block on the list - since we only initialize [ClientBlockList]
     /// when we're sure, we're about to add new elements to it, it always should
     /// stay non-empty.
-    pub(crate) fn last(&self) -> &Block {
+    pub(crate) fn last(&self) -> BlockPtr {
         self.get(self.integrated_len - 1)
     }
 
-    /// Returns a mutable block reference, given a pointer identifier of that block.
-    /// Returns `None` if no block with such reference could be found.
-    pub(crate) fn find(&mut self, ptr: &BlockPtr) -> Option<&mut Block> {
-        match self.try_get_mut(ptr.pivot()) {
-            Some(block) if block.contains(&ptr.id) => return Some(block),
-            _ => {
-                let pivot = self.find_pivot(ptr.id.clock)?;
-                self.try_get_mut(pivot)
-            }
-        }
-    }
-
     /// Given a block's identifier clock value, return an offset under which this block could be
     /// found using binary search algorithm.
     pub(crate) fn find_pivot(&self, clock: u32) -> Option<usize> {
@@ -279,21 +293,21 @@ impl ClientBlockList {
     /// list. Clocks are considered to work in left-side inclusive way, meaning that block with
     /// an ID (<client-id>, 0) and length 2, with contain all elements with clock values
     /// corresponding to {0,1} but not 2.
-    pub(crate) fn find_block(&self, clock: u32) -> Option<&Block> {
+    pub(crate) fn get_block(&self, clock: u32) -> Option<BlockPtr> {
         let idx = self.find_pivot(clock)?;
         self.try_get(idx)
     }
 
     /// Pushes a new block at the end of this block list.
-    pub(crate) fn push(&mut self, block: block::Block) {
-        self.list.push(UnsafeCell::new(block));
+    pub(crate) fn push(&mut self, block: Box<block::Block>) {
+        self.list.push(block);
         self.integrated_len += 1;
     }
 
     /// Inserts a new block at a given `index` position within this block list. This method may
     /// panic if `index` is greater than a length of the list.
-    fn insert(&mut self, index: usize, block: block::Block) {
-        self.list.insert(index, UnsafeCell::new(block));
+    fn insert(&mut self, index: usize, block: Box<block::Block>) {
+        self.list.insert(index, block);
         self.integrated_len += 1;
     }
 
@@ -319,12 +333,11 @@ impl ClientBlockList {
     pub(crate) fn squash_left(&mut self, index: usize) -> Option<SquashResult> {
         let replacement = {
             let (l, r) = self.list.split_at_mut(index);
-            let left = unsafe { &mut *l[index - 1].get() };
-            let right = unsafe { &*r[0].get() };
-            if left.is_deleted() == right.is_deleted() && left.same_type(right) {
+            let mut left = BlockPtr::from(&mut l[index - 1]);
+            let right = BlockPtr::from(&r[0]);
+            if left.is_deleted() == right.is_deleted() && left.same_type(right.deref()) {
                 if left.try_squash(right) {
-                    let new_ptr = BlockPtr::new(left.id().clone(), index as u32 - 1);
-                    Some(new_ptr)
+                    Some(left)
                 } else {
                     None
                 }
@@ -334,9 +347,9 @@ impl ClientBlockList {
         };
 
         if let Some(replacement) = replacement {
-            let block = self.list.remove(index).into_inner();
+            let block = self.list.remove(index);
             self.integrated_len -= 1;
-            if let Block::Item(item) = block {
+            if let Block::Item(item) = *block {
                 return Some(SquashResult {
                     parent: item.parent,
                     parent_sub: item.parent_sub,
@@ -370,22 +383,14 @@ impl Default for ClientBlockList {
     }
 }
 
-impl Index<usize> for ClientBlockList {
-    type Output = block::Block;
-
-    fn index(&self, index: usize) -> &Self::Output {
-        self.get(index)
-    }
-}
-
-pub(crate) struct ClientBlockListIter<'a>(std::slice::Iter<'a, UnsafeCell<block::Block>>);
+pub(crate) struct ClientBlockListIter<'a>(std::slice::Iter<'a, Box<block::Block>>);
 
 impl<'a> Iterator for ClientBlockListIter<'a> {
     type Item = &'a Block;
 
     fn next(&mut self) -> Option<Self::Item> {
         let next = self.0.next()?;
-        Some(unsafe { &*next.get() })
+        Some(next.as_ref())
     }
 }
 
@@ -420,6 +425,14 @@ impl BlockStore {
         self.clients.is_empty()
     }
 
+    pub fn contains(&self, id: &ID) -> bool {
+        if let Some(clients) = self.clients.get(&id.client) {
+            id.clock <= clients.last().last_id().clock
+        } else {
+            false
+        }
+    }
+
     /// Returns an immutable reference to a block list for a particular `client`. Returns `None` if
     /// no block list existed for provided `client` in current block store.
     pub fn get(&self, client: &u64) -> Option<&ClientBlockList> {
@@ -437,6 +450,10 @@ impl BlockStore {
         self.clients.iter()
     }
 
+    pub(crate) fn blocks(&self) -> Blocks<'_> {
+        Blocks::new(self)
+    }
+
     /// Returns a state vector, which is a compact representation of the state of blocks integrated
     /// into a current block store. This state vector can later be encoded and send to a remote
     /// peers in order to calculate differences between two stored and produce a compact update,
@@ -445,48 +462,35 @@ impl BlockStore {
         StateVector::from(self)
     }
 
-    /// Returns mutable reference to an item, given its pointer. Returns `None` if not such block
-    /// could be found.
-    pub(crate) fn get_item_mut(&self, ptr: &block::BlockPtr) -> Option<&mut block::Item> {
-        let block = self.get_block_mut(ptr)?;
-        block.as_item_mut()
-    }
-
     /// Returns immutable reference to a block, given its pointer. Returns `None` if not such
     /// block could be found.
-    pub(crate) fn get_block(&self, ptr: &block::BlockPtr) -> Option<&block::Block> {
-        let clients = self.clients.get(&ptr.id.client)?;
-        if let Some(block) = clients.try_get(ptr.pivot()) {
-            if block.contains(&ptr.id) {
-                return Some(&*block);
-            }
-        }
-        // ptr.pivot missed - go slow path to find it
-        let pivot = clients.find_pivot(ptr.id.clock)?;
-        ptr.fix_pivot(pivot as u32);
+    pub(crate) fn get_block(&self, id: &ID) -> Option<BlockPtr> {
+        let clients = self.clients.get(&id.client)?;
+        let pivot = clients.find_pivot(id.clock)?;
         clients.try_get(pivot)
     }
 
-    /// Returns immutable reference to a block, given its pointer. Returns `None` if not such
-    /// block could be found.
-    pub(crate) fn get_block_mut(&self, ptr: &block::BlockPtr) -> Option<&mut block::Block> {
-        let clients = self.clients.get(&ptr.id.client)?;
-        if let Some(block) = clients.try_get_mut(ptr.pivot()) {
-            if block.contains(&ptr.id) {
-                return Some(&mut *block);
-            }
+    pub(crate) fn get_item_clean_start(&mut self, id: &ID) -> Option<BlockPtr> {
+        let blocks = self.clients.get_mut(&id.client)?;
+        let mut index = blocks.find_pivot(id.clock)?;
+        let mut ptr = blocks.get(index);
+        if let Some(new) = ptr.splice(id.clock - ptr.id().clock) {
+            blocks.insert(index + 1, new);
+            index += 1;
         }
-        // ptr.pivot missed - go slow path to find it
-        let pivot = clients.find_pivot(ptr.id.clock)?;
-        ptr.fix_pivot(pivot as u32);
-        clients.try_get_mut(pivot)
+        Some(blocks.get(index))
     }
 
-    /// Returns immutable reference to an item, given its pointer. Returns `None` if not such
-    /// block could be found.
-    pub(crate) fn get_item(&self, ptr: &block::BlockPtr) -> Option<&block::Item> {
-        let block = self.get_block(ptr)?;
-        block.as_item()
+    pub(crate) fn get_item_clean_end(&mut self, id: &ID) -> Option<BlockPtr> {
+        let blocks = self.clients.get_mut(&id.client)?;
+        let index = blocks.find_pivot(id.clock)?;
+        let mut block = blocks.get(index);
+        let block_id = block.id();
+        if id.clock != block_id.clock + block.len() - 1 {
+            let new = block.splice(1 + id.clock - block_id.clock).unwrap();
+            blocks.insert(index + 1, new);
+        }
+        Some(BlockPtr::from(block))
     }
 
     /// Returns the last observed clock sequence number for a given `client`. This is exclusive
@@ -520,71 +524,17 @@ impl BlockStore {
             .or_insert_with(|| ClientBlockList::with_capacity(capacity))
     }
 
-    /// Given block pointer, tries to split it, returning a pointers to left and right halves
-    /// of a newly split block.
-    ///
-    /// If split was not necessary (eg. because block `ptr` was not inside of any block),
-    /// the right half returned wll be None.
-    ///
-    /// If no block for given `ptr` was found, then both returned options will be None.
-    pub fn split_block(&mut self, ptr: &BlockPtr) -> (Option<BlockPtr>, Option<BlockPtr>) {
-        let mut pivot = ptr.pivot();
-        if let Some(mut blocks) = self.clients.get_mut(&ptr.id.client) {
-            let block: &mut Block = {
-                let found = blocks.try_get_mut(pivot).and_then(|b| {
-                    if ptr.id.clock >= b.id().clock && ptr.id.clock < b.clock_end() {
-                        Some(&mut *b)
-                    } else {
-                        None
-                    }
-                });
-                if let Some(block) = found {
-                    block
-                } else {
-                    // search by pivot missed: perform standard lookup to find correct block
-                    if let Some(p) = blocks.find_pivot(ptr.id.clock) {
-                        pivot = p;
-                        blocks.get_mut(pivot)
-                    } else {
-                        return (None, None);
-                    }
-                }
-            };
+    /// Given block pointer, tries to split it, returning a true, if block was split in result of
+    /// calling this action, and false otherwise.
+    pub fn split_block(&mut self, mut block: BlockPtr, offset: u32) -> Option<BlockPtr> {
+        let id = block.id().clone();
+        let blocks = self.clients.get_mut(&id.client)?;
+        let index = blocks.find_pivot(id.clock)?;
+        let mut right = block.splice(offset)?;
+        let right_ptr = BlockPtr::from(&mut right);
+        blocks.insert(index + 1, right);
 
-            let left_split_ptr = BlockPtr::new(block.id().clone(), pivot as u32);
-            let right_split_ptr = match block {
-                Block::Item(item) => {
-                    let len = item.len();
-                    if ptr.id.clock > item.id.clock && ptr.id.clock <= item.id.clock + len {
-                        let index = pivot + 1;
-                        let diff = ptr.id.clock - item.id.clock;
-                        let right_split = item.split(diff);
-                        let right_split_id = right_split.id.clone();
-                        let right_ptr = right_split.right.clone();
-                        if let Some(right_ptr) = right_ptr {
-                            blocks = if right_ptr.id.client == ptr.id.client {
-                                blocks
-                            } else {
-                                self.clients.get_mut(&right_ptr.id.client).unwrap()
-                            };
-                            let right = blocks.find(&right_ptr).unwrap();
-                            if let Some(right_item) = right.as_item_mut() {
-                                right_item.left =
-                                    Some(BlockPtr::new(right_split.id.clone(), index as u32));
-                            }
-                        }
-                        blocks.insert(index, Block::Item(right_split));
-                        Some(BlockPtr::new(right_split_id, index as u32))
-                    } else {
-                        None
-                    }
-                }
-                _ => None,
-            };
-            (Some(left_split_ptr), right_split_ptr)
-        } else {
-            (None, None)
-        }
+        Some(right_ptr)
     }
 }
 
@@ -594,8 +544,8 @@ impl std::fmt::Display for ClientBlockList {
         let mut i = 0;
         writeln!(f, "")?;
         while i < self.list.len() {
-            let block = &self[i];
-            writeln!(f, "\t\t{}", block)?;
+            let block = self.get(i);
+            writeln!(f, "\t\t{}", block.deref())?;
             if i == self.integrated_len {
                 writeln!(f, "---")?;
             }
@@ -614,3 +564,44 @@ impl std::fmt::Display for BlockStore {
         writeln!(f, "}}")
     }
 }
+
+pub(crate) struct Blocks<'a> {
+    current_client: std::vec::IntoIter<(&'a u64, &'a ClientBlockList)>,
+    current_block: Option<ClientBlockListIter<'a>>,
+}
+
+impl<'a> Blocks<'a> {
+    fn new(update: &'a BlockStore) -> Self {
+        let mut client_blocks: Vec<(&'a u64, &'a ClientBlockList)> =
+            update.clients.iter().collect();
+        // sorting to return higher client ids first
+        client_blocks.sort_by(|a, b| b.0.cmp(a.0));
+        let mut current_client = client_blocks.into_iter();
+
+        let current_block = current_client.next().map(|(_, v)| v.iter());
+        Blocks {
+            current_client,
+            current_block,
+        }
+    }
+}
+
+impl<'a> Iterator for Blocks<'a> {
+    type Item = &'a Block;
+
+    fn next(&mut self) -> Option<Self::Item> {
+        if let Some(blocks) = self.current_block.as_mut() {
+            let block = blocks.next();
+            if block.is_some() {
+                return block;
+            }
+        }
+
+        if let Some(entry) = self.current_client.next() {
+            self.current_block = Some(entry.1.iter());
+            self.next()
+        } else {
+            None
+        }
+    }
+}
